/**
 *
 * Copyright (C) 2013 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/
#include "Cylinder.h"
#include <cmath>
// Constructors/Destructors
//  

Cylinder::Cylinder (const Parameters& p) : Node(p) {}

Cylinder::Cylinder (const Parameters& p, const Vector& a, const Vector& b , double r) : Node(p) {
  this->a = a;
  this->b = b;
  this->r = r;
  this->axe = Vector(1/(b[0]-a[0]),1/(b[1]-a[1]),1/(b[2]-a[2]));
  this->height = Norm(axe);
  this->center = (a-b)/2;
  this->p.setPosition(center);
}

Cylinder::Cylinder (const Parameters& p, const Vector& center, const Vector& axe , double height, double r) : Node(p) {
  this->a = center - axe*height/2.0;
  this->b = center + axe*height/2.0;
  this->r = r;
  this->axe = axe;
  this->height = height;
  this->center = center;
}


Cylinder::Cylinder (const Cylinder& c ) : Node(c.p){
  this->a = c.a;
  this->b = c.b;
  this->r = c.r;
  this->axe = c.axe;
  this->height = c.height;
  this->center = c.center;
}


Cylinder::~Cylinder ( ) { }

//  
// Methods
//  



int Cylinder::Inside (const Node&) const{
  return 1;
}


//x*x + y*y - r*r
int Cylinder::Intersect(const Ray& ray, Intersections& l) const{

  Vector u = Normalized(b-a);
  
  Vector dir=ray.Direction()/u;
  Vector orig= (ray.Origin()-a) / u;
  double A = dir[0] * dir[0] + dir[1] * dir[1] + dir[2] * dir[2];
  double B = orig[0]*(-dir[0]) + orig[1]*(-dir[1]) + orig[2]*(-dir[2]);
  double C = orig[0]*orig[0] + orig[1]*orig[1] +  orig[2]*orig[2] ;
  double delta = B*B - A*C + A*r*r;
  int result = 0;


  if (delta > 0.0){
    double d = sqrt(delta);
    double t2 = (B + d)/A;
    double t1 = (B - d)/A;
    //printf("A : t1=%f, t2=%f\n",t1,t2);
    if (t2 > 0.0)
      { 
	if(t1>1e-4){
	  Vector pt = ray(t1);
	  if( ( pt - b ) * u <= 0.0 && ( pt - a ) * u >= 0.0 ) {
	    Vector norm = Normalized( pt - (a + u * ( ( pt - a ) * u ) ) );
	    l.append(Parameters(p.getEmission(),p.getTexture(),p.getReflection(),Normalized(norm),pt, t1));
	    result = 1;
	  }
	}

	if(t2>1e-4){
	  Vector pt = ray(t2);
	  if( ( pt - b ) * u <= 0.0 && ( pt - a ) * u >= 0.0 ){
	    Vector norm = Normalized( pt - (a + u * ( ( pt - a ) * u ) ) );
	    l.append(Parameters(p.getEmission(),p.getTexture(),p.getReflection(),Normalized(norm),pt, t2));
	    result = 1;
	  }
	}
      }
    if(result==0){ //TODO
    }

  }
  return result;
}

int Cylinder::Inside (const Vector& p )const{
  double l = (p-a) * Normalized( b - a ); 
  if( l >= 0.0 && l <= this->height ) {
    double d = (p-b)*(p-b)-l*l;
    if( r*r < d ) return 1;
  }
  return 0;
}


// Accessor methods
//  


// Other methods
//  

